# Copyright 2022 Datum Technology Corporation
# SPDX-License-Identifier: Apache-2.0 WITH SHL-2.1
########################################################################################################################


version_text = "Moore.io CLI Client v1.2.6"
help_text = """
                              ███╗   ███╗ ██████╗  ██████╗ ██████╗ ███████╗   ██╗ ██████╗
                              ████╗ ████║██╔═══██╗██╔═══██╗██╔══██╗██╔════╝   ██║██╔═══██╗
                              ██╔████╔██║██║   ██║██║   ██║██████╔╝█████╗     ██║██║   ██║
                              ██║╚██╔╝██║██║   ██║██║   ██║██╔══██╗██╔══╝     ██║██║   ██║
                              ██║ ╚═╝ ██║╚██████╔╝╚██████╔╝██║  ██║███████╗██╗██║╚██████╔╝
                              ╚═╝     ╚═╝ ╚═════╝  ╚═════╝ ╚═╝  ╚═╝╚══════╝╚═╝╚═╝ ╚═════╝
                                 Moore.io (`mio`) Command Line Interface (CLI) - v1.2.6
                             User Manual: https://readthedocs.org/projects/mooreio-client/

              https://mooreio.org - Copyright 2021-2022 Datum Technology Corporation - https://datumtc.ca
Usage:
  mio sim        <ip>  [-t <test_name>]  [-s <seed>]  [-v <level>]  [-w]  [options]  [-- <args>]
  mio regr       <ip>  <regression> [-d | --dry]
  mio new
  mio install    <ip>  [-u <username>]  [-p <password>]
  mio init       [-p <path>]
  mio results    <ip> <filename>
  mio dox        <ip>
  mio cov        <ip>
  mio clean      <ip>  [-d | --deep]
  mio (-h | --help)
  mio --version

Options:
  -t <test_name>  Specify the UVM test to be run.  This argument is fed into 'test-name-template' from the target ip.yml.
  -s <seed>       Specify the seed for constrained-random testing.  If none is provided, a random one will be picked.
  -v <level>      Specify the UVM verbosity level for logging: none, low, medium, high or debug.  [default: medium]
  -w              Enable wave capture to disk.
  -c, --cov       Enable code & functional coverage capture.
  -g, --gui       Invoke the simulator's Graphical User Interface.
  ->              All UVM simulation logs are also output in JSON.
  --viv           Force mio to use Xilinx   Vivado      for simulation
  --mtr           Force mio to use Metrics  Simulator   for simulation
  --vcs           Force mio to use Synopsys VCS         for simulation
  --xcl           Force mio to use Cadence  Xcelium     for simulation
  --qst           Force mio to use Mentor   QuestaSim   for simulation
  --riv           Force mio to use Aldec    Riviera PRO for simulation
  -- <args>       Add arguments for compilation (+define+NAME[=VALUE]) or simulation (+NAME[=VALUE])).
                  NOTE: List must be within quotes ex: "+define+A +define+B=2 +ABC=123"
  -S              Force mio to simulate  target IP.  Can be combined with -C and/or -E.
  -E              Force mio to elaborate target IP.  Can be combined with -C and/or -S.
  -C              Force mio to compile   target IP.  Can be combined with -E and/or -S.
  -F              Force mio to invoke FuseSoC on core file(s).
  -d, --deep      Also cleans compiled dependencies.
  -d, --dry       Dry-run for regressions.  Compiles & elaborates as normal, but only prints out the tests it would run.
  --dbg           Prints tracing data
  -h --help       Show this screen.
  --version       Show version.
   
Examples:
  mio init                 # Create a new empty Moore.io project/IP in this location.
                           # *Highly* recommended for importing code to the ecosystem.
  mio new                  # Invoke the Moore.io UVM Framework Code Generator - 17 Templates, from advanced agents to tests.
  mio install uvmt_my_ip   # Install IP dependencies for 'uvmt_my_ip' from the Moore.io IP Marketplace: https://mooreio.org.

  mio sim     uvmt_my_ip -t smoke -s 1 -w -c           # Compile, elaborate and simulate test 'uvmt_my_ip_smoke_test_c'
                                                       # for IP 'uvmt_my_ip' with seed '1' and waves & coverage capture enabled.
  mio sim     uvmt_my_ip -t smoke -s 1 -- "+NPKTS=10"  # Compile, elaborate and simulate test 'uvmt_my_ip_smoke_test_c'
                                                       # for IP 'uvmt_my_ip' with seed '1' and waves and a simulation argument.
  mio sim -S  uvmt_my_ip -t smoke -s 42 -v high -g     # Only simulates test 'uvmt_my_ip_smoke_test_c' for IP 'uvmt_my_ip'
                                                       # with seed '42' and UVM_HIGH verbosity using the simulator in GUI mode.
  mio sim -C  uvmt_my_ip                               # Only compile 'uvmt_my_ip'.
  mio sim -E  uvmt_my_ip                               # Only elaborate 'uvmt_my_ip'.
  mio sim -CE uvmt_my_ip                               # Compile and elaborate 'uvmt_my_ip'.
  
  mio regr uvmt_my_ip sanity            # Run sanity regression for IP 'uvm_my_ip', from test suite 'ts.yml'
  mio regr uvmt_my_ip apb_xc.sanity     # Run sanity regression for IP 'uvm_my_ip', from test suite 'apb_xc.ts.yml'
  mio regr uvmt_my_ip axi_xc.sanity -d  # Dry-run sanity regression for IP 'uvm_my_ip', from test suite 'axi_xc.ts.yml'

  mio clean     uvmt_my_ip   # Delete compilation, elaboration and simulation binaries for IP 'uvmt_my_ip'
  mio results   uvmt_my_ip   # Parse simulation results and generates reports from an IP 'uvmt_my_ip's simulations
  mio cov       uvmt_my_ip   # Merge coverage data and generates reports from an IP 'uvmt_my_ip's simulations
  mio dox       uvmt_my_ip   # Invoke Doxygen to generates reference documentation for IP 'uvmt_my_ip'
  
"""


from mio import cfg
from mio import clean
from mio import cov
from mio import dox
from mio import results
from mio import sim
from mio import regr
from mio import install
from mio import cache
from mio import new
from mio import common
from mio import init
from mio import main

import sys
import argparse
import os
import subprocess
import shutil
import yaml
from datetime import datetime
from yaml.loader import SafeLoader
import re
import random
import pathlib


uvm_levels = ["none","low","medium","high","debug"]
simulators = ["viv","mtr","vcs","xcl","qst","riv"]


def main():
    parser       = build_parser()
    cli_args     = parser.parse_args()
    cfg.dbg      = cli_args.dbg
    common.dbg("CLI arguments: " + str(cli_args))
    cfg.cli_args = cli_args
    
    if cli_args.help:
        print_help_text()
    if cli_args.version:
        print_version()
    
    if cli_args.wd != None:
        cfg.set_pwd(cli_args.wd)
    else:
        cfg.pwd = os.getcwd()
    in_project = cache.find_project_descriptor()
    
    if cli_args.command == "init":
        if cli_args.path != None:
            cfg.set_pwd(cli_args.path)
    if not in_project:
        if cli_args.command != "init":
            common.fatal("Could not find 'mio.toml' project file")
        else:
            init.new_project()
            common.exit()
    
    cache.load_tree()
    cache.load_configuration()
    common.create_common_files()
    log_cli_args_to_disk()
    cache.scan_and_load_ip_metadata()
    
    if cli_args.command == 'init':
        init.new_ip()
        common.exit()
    if cli_args.command == 'new':
        new.menu(cfg.cli_args.template)
        common.exit()
    if cli_args.command == 'install':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        install.install_all_ips_for_target(cli_args.is_global, cli_args.username, cli_args.password)
        common.exit()
    if cli_args.command == 'dox':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        dox.gen_doxygen()
        common.exit()
    if cli_args.command == 'cov':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        cov.gen_cov_report()
        common.exit()
    if cli_args.command == 'results':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        results.do_parse_results()
        common.exit()
    if cli_args.command == 'clean':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        clean.do_clean()
        common.exit()
    if cli_args.command == 'sim':
        cfg.target_ip_name = cli_args.ip
        sim_args = convert_cli_to_sim_args(cli_args)
        cache.check_target_ip()
        sim.main(sim_args)
        common.exit()
    if cli_args.command == 'regr':
        cfg.target_ip_name = cli_args.ip
        cache.check_target_ip()
        regr.main()
        common.exit()
    
    common.fatal("No command specified")


def build_parser():
    parser = argparse.ArgumentParser(prog="mio", description="", add_help=False)
    parser.add_argument("-h"   , "--help"   , help="Show this help message and exit.", action="store_true", default=False, required=False)
    parser.add_argument("-v"   , "--version", help="Print the mio version and exit." , action="store_true", default=False, required=False)
    parser.add_argument("--dbg",              help="Enable mio tracing output."      , action="store_true", default=False, required=False)
    parser.add_argument("-C"   , "--wd"     , help="Run as if mio was started in <path> instead of the current working directory.", type=pathlib.Path, required=False)
    subparsers = parser.add_subparsers(help='Command to be performed by mio', dest='command')
    
    parser_init = subparsers.add_parser('init', help='Create a new empty Moore.io project/IP in this location.')
    parser_init.add_argument('-p', "--path", help='Working directory path', type=pathlib.Path, required=False)
    
    parser_new = subparsers.add_parser('new', help='Invoke the Moore.io UVM Framework Code Generator - 17 Templates, from advanced agents to tests.')
    parser_new.add_argument('-t', '--template', help='Template name', choices=new.templates, required=False)
    
    parser_install = subparsers.add_parser('install', help='Install IP dependencies from the Moore.io IP Marketplace: https://mooreio.org.')
    parser_install.add_argument('ip'              , help='Target IP'                                       )
    parser_install.add_argument("-g", "--global"  , help="Install dependencies under '~/.mio'." , action="store_true", default=False, required=False, dest="is_global")
    parser_install.add_argument('-u', "--username", help='Moore.io IP Marketplace username', required=False)
    parser_install.add_argument('-p', "--password", help='Moore.io IP Marketplace password', required=False)
    
    parser_sim = subparsers.add_parser('sim', help='Compile, elaborate and simulate an IP.')
    parser_sim.add_argument('ip'               , help='Target IP'                                                                                                                               )
    parser_sim.add_argument('-t', "--test"     , help='Delete compiled IP dependencies.'                                                                                        , required=False)
    parser_sim.add_argument('-s', "--seed"     , help='Specify the seed for constrained-random testing.  If none is provided, a random one will be picked.', type=int           , required=False)
    parser_sim.add_argument('-v', "--verbosity", help='Specify the UVM verbosity level for logging: none, low, medium, high or debug.  Default: medium'    , choices=uvm_levels , required=False)
    parser_sim.add_argument('-w', "--waves"    , help='Enable wave capture to disk.'                                                                       , action="store_true", required=False)
    parser_sim.add_argument('-c', "--cov"      , help='Enable code & functional coverage capture.'                                                         , action="store_true", required=False)
    parser_sim.add_argument('-g', "--gui"      , help="Invoke the simulator's Graphical User Interface."                                                   , action="store_true", required=False)
    #parser_sim.add_argument('-p', "--prism"    , help='Enable Moore.io PRISM advanced UVM debugging.'                                                      , action="store_true", required=False)
    parser_sim.add_argument('-a', "--app"      , help='Specifies which simulator to use: viv, mtr, vcs, xcl, qst, riv.'                                    , choices=simulators , required=False)
    parser_sim.add_argument('-S'               , help='Force mio to simulate target IP.  Can be combined with -F, -C and/or -E.'                           , action="store_true", required=False)
    parser_sim.add_argument('-E'               , help='Force mio to elaborate target IP.  Can be combined with -F, -C and/or -S.'                          , action="store_true", required=False)
    parser_sim.add_argument('-C'               , help='Force mio to compile target IP.  Can be combined with -F, -E and/or -S.'                            , action="store_true", required=False)
    parser_sim.add_argument('-F'               , help='Force mio to invoke FuseSoC on core file(s).  Can be combined with -C, -E and/or -S.'               , action="store_true", required=False)
    parser_sim.add_argument('--'               , help='Add arguments for compilation (+define+NAME[=VALUE]) or simulation (+NAME[=VALUE])).'               , dest='add_args'    , required=False)
    
    parser_sim = subparsers.add_parser('regr', help='Run a regression on an IP.')
    parser_sim.add_argument('ip'         , help='Target IP'            )
    parser_sim.add_argument('regr'       , help='Regression to be run.  For Test Bench IPs with multiple Test Suites, the suite must be specified. Ex: `mio regr my_ip apbxc.sanity`')
    parser_sim.add_argument('-a', "--app", help='Specifies which simulator to use: viv, mtr, vcs, xcl, qst, riv.', choices=simulators , required=False)
    parser_sim.add_argument('-d', "--dry", help='Compiles and elaborates target IP but only prints out the tests that would be run.', action="store_true", default=False , required=False)
    
    parser_clean = subparsers.add_parser('clean', help='Delete compilation, elaboration and simulation artifacs for an IP')
    parser_clean.add_argument('ip'          , help='Target IP'                                                            )
    parser_clean.add_argument('-d', "--deep", help='Delete compiled IP external dependencies.', action="store_true", required=False)
    
    parser_results = subparsers.add_parser('results', help='Parse simulation results and generates reports for an IP.')
    parser_results.add_argument('ip'      , help='Target IP'      )
    parser_results.add_argument('filename', help='Report filename')
    
    parser_cov = subparsers.add_parser('cov', help='Merge simulation coverage data and generate reports for an IP.')
    parser_cov.add_argument('ip', help='Target IP')
    
    parser_dox = subparsers.add_parser('dox', help='Invoke Doxygen to generate reference documentation for an IP.')
    parser_dox.add_argument('ip', help='Target IP')
    
    return parser


def convert_cli_to_sim_args(cli_args):
    args = {}
    
    if not cli_args.C and not cli_args.E and not cli_args.S:
        args['cmp' ] = True
        args['elab'] = True
        args['sim' ] = True
    else:
        if cli_args.C:
            args['cmp'] = True
        else:
            args['cmp'] = False
        if cli_args.E:
            args['elab'] = True
        else:
            args['elab'] = False
        if cli_args.S:
            args['sim'] = True
        else:
            args['sim'] = False
    if cli_args.seed == None:
        args['seed'] = random.randint(1, 2147483646)
        common.dbg("Seed not specified, picked " + str(args['seed']))
    else:
        args['seed'] = cli_args.seed
    args['test_name'] = cli_args.test
    args['args'] = cli_args.add_args
    if cli_args.verbosity == None:
       args['verbosity'] = "medium"
    else:
       args['verbosity'] = cli_args.verbosity
    args['waves'] = cli_args.waves
    args['cov'] = cli_args.cov
    args['gui'] = cli_args.gui
    if cli_args.gui:
        args['waves'] = False
    
    if cli_args.app != None:
        if cli_args.app == "viv":
            args['simulator'] = "vivado"
        elif cli_args.app == "mtr":
            args['simulator'] = "mtr"
        elif cli_args.app == "vcs":
            args['simulator'] = "vcs"
        elif cli_args.app == "xcl":
            args['simulator'] = "xcelium"
        elif cli_args.app == "qst":
            args['simulator'] = "questa"
        elif cli_args.app == "riv":
            args['simulator'] = "riviera"
        else:
            common.dbg("Picked default simulator: vivado")
            args['simulator'] = "vivado"
    else:
        common.dbg("Picked default simulator: vivado")
        args['simulator'] = "vivado"
    common.dbg("Using simulator " + args['simulator'])
    
    return args


def print_help_text():
    print(help_text)
    common.exit(False)


def print_version():
    print(version_text)
    common.exit(False)


def log_cli_args_to_disk():
    try:
        common.create_file(cfg.commands_file_path)
        with open(cfg.commands_file_path, 'r') as yaml_file_read:
            ymlr = yaml.load(yaml_file_read, Loader=SafeLoader)
            if not ymlr:
                ymlr = {}
            ymlr[common.timestamp()] = sys.argv
            with open(cfg.commands_file_path, 'w') as yaml_file_write:
                yaml.dump(ymlr, yaml_file_write)
    except Exception as e:
        common.warning("Failed to log command history to disk: " + str(e))
